/** 
 * Rewriting module
 * This module rewrites an expression using an eventually outermost strategy
 * The rewriting rules are:
 *   - Axioms given by the user (left to right)
 *   - if TRUE then E1 else E2 fi -> E1
 *   - if FALSE then E1 else E2 fi -> E2
 *   - Si E1:T o E2:T, EQUAL(E1,E2) -> EQ_T(E1,E2)
 *
 * @author: Vladimir Tamara
 * @version: 2.0
 **/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "definic.h"
#include "os_dep.h"
#include "apimanta.h"

/* GLOBAL DATA */

nodo_igualdad   *ifrules=NULL, /* if rules */
                *eqrules=NULL; /* EQUAL rules */
                
/**
 * Creates an expression from a string
 * @author Vladimir Tamara  
 **/
int                                      /* 0 iff everything is fine */
expressionsOfString(char *s,             /* String to parse */
		    nodo_argum **e,      /* Resulting expressions */ 
		    char *merr)          /* Errors buffer */
{
  FILE *fp;
  cad50 arcnam;
  texpr mens_error;

  assert(s!=NULL);
  assert(e!=NULL);

  if (*e!=NULL)
    { elimLArgumento(e); }

  strcpy(arcnam,"normaliz");
  fp=fopen(arcnam,"wt");
  if (fp==NULL)
    {
      sprintf(mens_error,mensajes[196],arcnam);
      strcat(mens_error,"\n");
      strcat(merr,mens_error);
      return -1;
    }
  fprintf(fp,"%s",s);
  fclose(fp);

  fp=fopen(arcnam,"rt");
  if (fp==NULL)
    {
      sprintf(mens_error,mensajes[196],arcnam);
      strcat(mens_error,"\n");
      strcat(merr,mens_error);
      return -1;
    }
  *e = analice_lista_exp (fp, mens_error);
  if (*e==NULL)
    {
      strcat(merr,mens_error);
      strcat(merr,"\n");
      return -1;
    }
  fclose(fp);

  unlink("normaliz");
  return 0;
} /* expressionsOfString */



/**
 * Creates an equality from two strings
 * @author Vladimir Tamara  
 **/
int                                   /* 0 iff everything is OK */
equalityOfStrings(char *li,           /* Left side */
		  char *ld,           /* Right side */
		  nodo_igualdad **eq, /* Returned equality */
		  char *merr)         /*Errors buffer */
{
  nodo_argum *ar;

  if (*eq!=NULL)
    { elimLIgualdad(eq); }

  /* Left Side */
  if (nuevoNodoIgualdad(eq,merr)!=0)
    { return -1; }
  strcpy((*eq)->li_str,li);
  ar=NULL;
  if (expressionsOfString(li,&ar,merr)!=0)
    { return -1; }
  assert(ar->sig==NULL); /* Just one expression at left side */
  (*eq)->lado_izq=ar->expr;
  ar->expr=NULL;
  elimLArgumento(&ar);

  /* Right side */
  strcpy((*eq)->ld_str,ld);
  if (expressionsOfString(ld,&ar,merr)!=0)
    { return -1; }
  assert(ar->sig==NULL); /* Just one expression at right side */
  (*eq)->lado_der=ar->expr;
  ar->expr=NULL;
  elimLArgumento(&ar);

  return 0;  
} /* equalityOfStrings */



/**
 * Init rewriting system 
 * @author Vladimir Tamara  
 **/
int                             /* 0 iff everything is OK */
init_rewriting(char *merr)      /* Buffer for error messages */
{
  char cad[100];
  /* RULES for IF , it is not strict */
  ifrules=NULL;
  sprintf(cad,"%s TRUE %s X %s Y %s",claves[8],claves[9],claves[10],
	  claves[15]);
  if (equalityOfStrings(cad,"X",&ifrules,merr)!=0)
    { return -1;}
  ifrules->sig=NULL;
  sprintf(cad,"%s FALSE %s X %s Y %s",claves[8],claves[9],claves[10],
	  claves[15]);
  if (equalityOfStrings(cad,"Y",
		       &(ifrules->sig),merr)!=0)
    { return -1;}
  ifrules->sig->sig=NULL;

  eqrules=NULL;
  if (equalityOfStrings("EQUAL(X,Y)","EQUAL(X,Y)",&eqrules,merr)!=0)
    { return -1;}
  eqrules->sig=NULL;

  return 0;
} /* init_rewriting */



/**
 * Ends rewriting system 
 * @author Vladimir Tamara  
 **/
void
end_rewriting(void)
{
  if (ifrules!=NULL)
    { elimLIgualdad(&ifrules); }
  if (eqrules!=NULL)
    { elimLIgualdad(&eqrules); }
} /* init_rewriting */



/**
 * Changes the subexpression expr and puts the expression newexp.
 * @author Vladimir Tamara  
 **/
void 
changeSubexpressionByExpression(
  nodo_expresion **entexpr,  /* Entire expression */
  nodo_expresion **expr,     /* Subexpression to replace */
  nodo_expresion *newexp)    /* New expression to put */
{
  assert(entexpr!=NULL);
  assert(expr!=NULL);

  if (*expr==*entexpr)
    {
      elimNodoExpresion(expr);
      *expr=*entexpr=newexp;
    }
  else
    {
      elimNodoExpresion(expr);
      *expr=newexp;
    }
} /* changeSubexpressionByExpression */



/**
 * Changes the subexpression expr and puts the subexpression subexp.
 * @author Vladimir Tamara  
 **/
void 
changeExpressionBySubexpression(
  nodo_expresion **expr,     /* Subexpression to replace */
  nodo_expresion **subexp)    /* New expression to put */
{
  nodo_expresion *aux;

  assert(expr!=NULL);
  assert(subexp!=NULL);
  /*PRE:  subexp is a subexpression of expr (in the same tree) */

  if (*expr==*subexp)
    { return; }
  aux=*subexp;
  *subexp=NULL;
  elimNodoExpresion(expr);
  *expr=aux;
} /* changeExpressionBySubexpression */



/**
 * Rewrites a subexpression trying to unify with  a given set of rules
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something  in appliedrule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in appliedrule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 26-Jun-1998
 **/
int
rewriteSubexpressionWithRules(
                    nodo_expresion **entexpr,     /* Entire expression */
		    nodo_expresion **expr,        /* Subexpression */
		    nodo_igualdad  *rules,        /* Rules to try */ 
		    nodo_igualdad  **appliedrule, /* Rule used */
		    char *merr)                   /* Errors buffer */
{
  short rewriting; /* Returned value */
  Ptr_tabla unif;
  nodo_expresion *expaux;

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(appliedrule!=NULL);
  assert(merr!=NULL);

  rewriting=0;
  *appliedrule=rules;
  while (*appliedrule!=NULL && rewriting==0)
    {
      unif = umg1 (lir (*appliedrule), *expr);
      if (unif != NULL)
	{
	  expaux = apl (ldr (*appliedrule), unif, 1, &rewriting);
	  elimLTabla(&unif);
	  if (expaux==NULL)
	    {
	      strcat(merr,mensajes[166]);
	      strcat(merr,"\n");
	      return -1;
	    }
	  rewriting=1;
	}
      else
	{
	  *appliedrule=(*appliedrule)->sig;
	}	  
    }
  if (rewriting==1)
    { changeSubexpressionByExpression(entexpr,expr,expaux);  }

  return (int)rewriting;
} /* rewriteSubExpressionWithRules */



/**
 * Rewrites an IF expression
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something in appliedrule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in appliedrule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 26-Nov-1998
 **/
int
rewriteSubexpressionIF(
                    nodo_expresion **entexpr,     /* Entire expression */
		    nodo_expresion **expr,        /* Subexpression */
		    nodo_igualdad  **appliedrule, /* Rule used */
		    char *merr)                   /* Errors buffer */
{
  short rewriting; /* Returned value */

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(strcmp_nocase((*expr)->nombre,"IF")==0);
  assert(merr!=NULL);
  assert(ifrules!=NULL); 
  assert(ifrules->sig!=NULL);

  rewriting=0;
  
  if (strcmp((*expr)->argumentos->expr->nombre,"TRUE")==0)
    {
      *appliedrule=ifrules;
      changeExpressionBySubexpression(expr,
				      &((*expr)->argumentos->sig->expr));
      rewriting=1;
    }
  else if (strcmp((*expr)->argumentos->expr->nombre,"FALSE")==0)
    {
      *appliedrule=ifrules->sig;
      changeExpressionBySubexpression(expr,
				      &((*expr)->argumentos->sig->sig->expr));
      rewriting=1;
    }

  return (int)rewriting;
} /* rewriteSubexpressionIF */



/**
 * Tries to rewrite a subexpression that is an EQUAL
 *
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something and in rule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in rule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *  
 * @author Vladimir Tamara  
 * @date 26-Nov-1998
 **/
int 
rewriteSubexpressionEQUAL(
			 nodo_expresion **entexpr, /* Entire expression */
			 nodo_expresion **expr,    /* Subexpression */
			 nodo_igualdad **rule,     /* Rewriting rule used */
			 char *merr)               /* Errors buffer */
{
  int rewriting; /* Returned value */

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(merr!=NULL);
  assert(rule!=NULL);
  /* PRE: expr is subexpression of entexpr */
  assert(*expr!=NULL);
  assert(strcmp((*expr)->nombre,"EQUAL")==0);
  assert((*expr)->argumentos!=NULL);
  assert((*expr)->argumentos->expr!=NULL);
  assert((*expr)->argumentos->sig!=NULL);
  assert((*expr)->argumentos->sig->expr!=NULL);
  assert(eqrules!=NULL);

  rewriting=0;
  if (((*expr)->argumentos->expr->tipo)[0]!='\0')
    {
      *rule=eqrules;
      sprintf((*expr)->nombre,"EQ_%s",(*expr)->argumentos->expr->tipo);
      strcpy((*rule)->lado_der->nombre,(*expr)->nombre);
      sprintf((*rule)->ld_str,"%s(X,Y)",(*expr)->nombre);
      rewriting=1;
    }
  else if (((*expr)->argumentos->sig->expr->tipo)[0]!='\0')
    {
      *rule=eqrules;
      sprintf((*expr)->nombre,"EQ_%s",(*expr)->argumentos->sig->expr->tipo);
      strcpy((*rule)->lado_der->nombre,(*expr)->nombre);
      sprintf((*rule)->ld_str,"%s(X,Y)",(*expr)->nombre);
      rewriting=1;
    }
  
  return rewriting;
} /* rewriteSubexpressionEQUAL */



/**
 * Rewrites the subexpression that receives with a rule and the unifier 
 *
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something in appliedrule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in rule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 26-Jun-1998
 **/
int 
rewriteExpressionWithOneAxiom(
                    nodo_expresion **entexpr, /* Entire expression */
		    nodo_expresion **expr,    /* Subexpression */
		    nodo_igualdad  *rule,     /* Rule (axiom) to try */ 
		    Ptr_tabla unif,           /* Substitution */
		    char *merr)               /* Errors buffers */
{
  short rewriting;
  nodo_expresion *expaux;

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(rule!=NULL);
  assert(unif!=NULL);
  assert(unif->entrada!=NULL);
  assert(merr!=NULL);

  rewriting=0;

  expaux = apl (ldr (rule), unif, 1, &rewriting);
  if (expaux==NULL)
    {
      strcat(merr,mensajes[166]);
      strcat(merr,"\n");
      return -1;
    }
  changeSubexpressionByExpression(entexpr,expr,expaux);
  rewriting=1;

  return (int)rewriting;
} /* rewriteExpressionWithOneAxiom */



/**
 * Rewrites a subexpression trying to unify with  a given set of axioms
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something in appliedrule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in rule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 26-Jun-1998
 **/
int
rewriteSubexpressionWithAxioms(
                    nodo_expresion **entexpr, /* Entire expression */
		    nodo_expresion **expr,    /* Subexpression */
		    nodo_igualdad  *rules,        /* Rules (axioms) to try */ 
		    nodo_igualdad  **appliedrule, /* Rule used */
		    char *merr)               /* Errors buffer */
{
  int rewriting; /* Returned value */
  Ptr_tabla unif;

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(appliedrule!=NULL);
  assert(merr!=NULL);

  *appliedrule=rules;
  while (*appliedrule!=NULL)
    {
      unif = umg1 (lir (*appliedrule), *expr);
      if (unif != NULL)
	{
	  rewriting=rewriteExpressionWithOneAxiom(entexpr,expr,
						  *appliedrule,unif,merr);
	  elimLTabla(&unif);
	  return rewriting;
	}
      else
	{
	  *appliedrule=(*appliedrule)->sig;
	}	  
    }
  return 0;
} /* rewriteSubexpressionWithAxioms */



/**
 * Rewrites a subexpression that is a selector 
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something in rule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in rule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 26-Jun-1998
 **/
int 
rewriteSubexpressionSelector(
                    nodo_expresion **entexpr, /* Entire expression */
		    nodo_expresion **expr,    /* Subexpression */
		    nodo_igualdad *special,   /* Special rules */
		    nodo_igualdad **rule,     /* Rewriting rule used */
		    char *merr)               /* Errors buffer */
{
  int rewriting; /* Returned value */
  cad50 ntad;
  nodo_tad *expadt;
  texpr mens_error;
  nodo_select *sel;
  
  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(merr!=NULL);
  assert(rule!=NULL);
  /* PRE: expr is subexpression of entexpr */
  assert(*expr!=NULL);
  
  /* First we try with special rules */
  rewriting=rewriteSubexpressionWithRules(entexpr,expr,special,rule,merr);

  /* Then we try with axioms for selector */ 
  if (rewriting==0 && 
      existsSymbol(symtable,(*expr)->nombre,ntad)==SYM_SELECTOR)
    {
      /* ADT of selector is left in expadt */
      if ((expadt=buscaNombreLTipo(tad,ntad))==NULL &&
	  busque_arch_tad(ntad)==0)
	{
	  sprintf(mens_error,mensajes[194],ntad);
	  strcat(merr,mens_error);
	  strcat(merr,"\n");
	  return -1;
	}
      if (expadt==NULL)
	{ expadt=buscaNombreLTipo(tad,ntad); }
      assert(expadt!=NULL); /* busque_arch_tad has bugs */

      /* Selector definition is left in sel */
      for (sel=expadt->selectoras;sel!=NULL && 
	     strcmp(sel->nombre,(*expr)->nombre)!=0;
	   sel=sel->sig);
      assert(sel!=NULL); /* Bad symbols table */
      rewriting=rewriteSubexpressionWithAxioms(entexpr,expr,sel->axiomas,
					      rule,merr);      
    }

  return rewriting;
}  /* rewriteSubexpressionSelector */



/**
 * Tries to rewrite a subexpression of an expression. 
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something in rule the rewriting rule used
 *  -1 when an error ocurrs
 *
 * The applied rule returned in rule will no be eliminated 
 * hence it can be a reference to an axiom o to a general rule
 *
 * @author Vladimir Tamara  
 * @date 25-Jun-1998
 **/
int 
rewriteSubexpression(
		     nodo_expresion **entexpr, /* Entire expression */
		     nodo_expresion **expr,    /* Subexpression */
		     nodo_igualdad *special,    /* Special rules */
		     nodo_igualdad **rule,     /* Rewriting rule used */
		     char *merr)               /* Errors buffer */
{
  int rewriting; /* Returned value */

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(merr!=NULL);
  assert(rule!=NULL);
  /* PRE: expr is subexpression of entexpr */
  
  if (*expr==NULL)
    { rewriting= 0;  }
  else if (strcmp_nocase((*expr)->clase,"VAR")==0 || 
	   strcmp((*expr)->nombre,"ERROR")==0)
    /* The normal form of variables and ERROR is the same */
    { rewriting= 0; }
  else if (strcmp_nocase((*expr)->nombre,claves[8])==0)
    /* If is an special case */
    {rewriting=rewriteSubexpressionIF(entexpr,expr,rule,merr);}
  else if (strcmp((*expr)->nombre,"EQUAL")==0)
    { rewriting=rewriteSubexpressionEQUAL(entexpr,expr,rule,merr);  }
  else if (strcmp_nocase((*expr)->clase,"INIC")==0 ||
	   strcmp_nocase((*expr)->clase,"CONSTR")==0)
    /* We can't rewrite generators */
    { rewriting=0; }
  else /* Is a selector or an undefined symbol */
    { rewriting=rewriteSubexpressionSelector(entexpr,expr,special,rule,merr); }

  return rewriting;
} /* rewriteSubexpression */



/**
 * Rewrites and expression using the leftmost rule but just
 * in a given tree depth
 * @return
 *  0 if it can't rewrite anything
 *  1 if it can rewrite something
 *  -1 when an error ocurrs
 *
 * @author Vladimir Tamara  
 * @date 25-Jun-1998
**/
int
rewriteOutermost_nth(
	     nodo_expresion ** entexpr, /* Entiere expression */
	     nodo_expresion ** expr,     /* Subexpression to normalice */
	     nodo_igualdad *special,     /* Special rules */
	     int withHistory,            /* true if history is desired */
	     nodo_historia ** hfirst,    /* First in history */
	     nodo_historia ** hlast,     /* Last in history */
	     int n,                      /* Depth to reach in recursion */
	     char *merr)                 /* Error messages */  
{

  int rewriting; /* Returned value */
  nodo_igualdad *rule=NULL; /* Rule used to rewrite */ 
  nodo_historia *aux; /* History */
  nodo_argum *ar;

  assert(entexpr!=NULL);
  assert(expr!=NULL);
  assert(merr!=NULL);
  assert(n>=0);
  
  rewriting=0;
  if (n==0) /* Try to unify with a rule in this level */
    { 
      if ((rewriting=rewriteSubexpression(entexpr,expr,
					  special,&rule,merr))==1 &&
	  withHistory)
	{
	  aux = crear_historia (*entexpr, rule, "");
	  if (aux==NULL)
	    {  return -1;  }
	  else if (*hlast==NULL)
	    {
	      *hfirst=*hlast=aux;
	      (*hlast)->sig=NULL;
	    }
	  else
	    {
	      (*hlast)->sig = aux;
	      (*hlast) = aux;
	    }
	}
      /* Rules must not be deleted (they are axioms or special) */
    }
  else /* Search deeper */
    { 
      if (strcmp_nocase((*expr)->nombre,"IF")==0)
	{
	  /* We don't have to check 2nd and 3rd argument just the first */
	  rewriting=rewriteOutermost_nth(entexpr,
					&((*expr)->argumentos->expr),special,
					withHistory,hfirst,hlast,n-1,merr);
	}
      else
	{
	  for (ar=(*expr)->argumentos;
	       ar!=NULL && 
		 (rewriting=rewriteOutermost_nth(entexpr,&(ar->expr),
						special,withHistory,
						hfirst,hlast,n-1,merr))==0;
	       ar=ar->sig);
	}
    }
  return rewriting;
} /* rewriteOutermost_nth */



/**
 * Returns the height of an expression's tree
 *
 * @author Vladimir Tamara  
 * @date 25-Jun-1998
 **/
int                                           /* Expression's height */ 
expressionHeight(nodo_expresion *e)           /* Expression to check */
{
  nodo_argum *ar;
  int heightSon,maxHeightSon;

  if (e==NULL)
    { return 0; }
  maxHeightSon=0;
  if (strcmp_nocase(e->nombre,"IF")==0)
    {
      /* Just the condition's height, because IF is not strict */
      maxHeightSon=expressionHeight(e->argumentos->expr);
    }
  else
    {
      for (ar=e->argumentos;ar!=NULL;ar=ar->sig)
	{
	  if ((heightSon=expressionHeight(ar->expr))>maxHeightSon)
	    { maxHeightSon=heightSon; }
	}
    }
  return maxHeightSon+1;
} /* expressionHeight */



/**
 * Rewrites and expression using an eventually outermost strategy
 * @return
 * 0 if it can't rewrite anything
 * 1 if it can rewrite something
 * -1 when an error ocurrs
 * 
 * @author Vladimir Tamara  
 * @date 25-Nov-1998
 **/
int              /* -1 if error otherwise amount of rewrite rules applied */
rewriteOutermost(nodo_expresion ** entexpr, /* Entire expression */
		 nodo_igualdad *special,     /* Special Rules */ 
		 int withHistory,            /* true if history is desired */
		 nodo_historia ** hfirst,    /* First in history */
		 nodo_historia ** hlast,     /* Last in history */
		 char *merr)
{
  int depth; /* Depth to reach in expression tree */
  int maxdepth; /* height of tree */
  int rewriting; /* 1 iff it could rewrite */
  int retval; /* Value to return */
  int attempts;  /* Attempts to rewrite */

  assert(entexpr!=NULL);
  assert(merr!=NULL);
  
  maxdepth=expressionHeight(*entexpr);
  depth=0;
  attempts=0;
  retval=0;
  while (depth<maxdepth && attempts<MAX_LEFTMOST_REWRITE)
    {
      attempts++;
      if ((rewriting=rewriteOutermost_nth(entexpr,entexpr,special,
				 withHistory,hfirst,hlast,depth,merr))==-1)
	/* An error ocurred */
	{ return -1; }
      else if (rewriting==0)
	/* It couldn't rewrite in level depth */
	{  depth++; }
      else
	/* It could rewrite something in level depth */
	{ 
	  depth=0; 
	  retval++;
	  maxdepth=expressionHeight(*entexpr);
	}
    }
  if (attempts>=MAX_LEFTMOST_REWRITE)
    {
      strcat(merr,mensajes[202]);
      strcat(merr,"\n");
      return -1;
    }
  return retval;
} /* rewriteOutermost */
