 
/** 
 * SYMBOL TABLE
 * Written by Vladimir Tamara
 * 
 * This module mantains a symbol table with all the defined symbols
 **/
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "definic.h"
#include "os_dep.h"
#include "apimanta.h"


nodo_tad *symtable;     /* Symbols table. VTP 4/May/1998 */

/**
 * Frees all the axioms of the selectors in the list that receives
 * but it doesn't free the selectors (it leaves the signature)
 *
 *@param ns List of selectors finished with NULL
 **/
void freeAxioms(nodo_select *ns)
{
  while (ns!=NULL)
    {
      elimLIgualdad(&(ns->axiomas));
      ns->axiomas=NULL;
      ns=ns->sig;
    }
}


/**
 * Loads a symbol table.  (Open all the ADT and representations and
 * stores information of each one)
 *
 * @param symt Symbol table
 * @param merr Buffer to add error messages
 **/
int loadSymbols(nodo_tad **symt,char *merr)
{

  nodo_nombre *nom,*auxn;
  FILE *fp;
  char *aname;
  nodo_tad **st;
  int rc=0;
  
  nom=NULL;
  aname=(char *)malloc(sizeof(char)*500);
  if (aname==NULL)
    {
      strcat(merr,mensajes[166]);
      strcat(merr,"\n");
      rc=-1;
      goto exit;
    }
  /* First we store all the ADT names */
  nom=lista_dir (direct[2],"adt",1);
  if (nom==NULL)
    {
      strcat(merr,mensajes[166]);
      strcat(merr,"\n");
      rc=-1;
      goto exit;
    }      

  /*Then we have to load each one and delete the axioms, because
    we just need the signature */
  for (*symt=NULL,st=symt,auxn=nom;auxn!=NULL;auxn=auxn->sig)
    {
      sprintf(aname,"%s/%s",direct[2],auxn->nombre);
      if (strcmp(aname+strlen(aname)-3,"adt")==0)
        {
          fp=fopen(aname,"rt");
          if (fp!=NULL)
            {
              if (cargaTipo(fp,st,merr)!=0)
                {
                  rc=-1;
                  goto exit;
                }
              freeAxioms((*st)->selectoras);
              st=&((*st)->sig);
	      fclose(fp);
            }
        }
    }
exit:
  if (aname!=NULL)
    {
      free(aname);
    }
  if (nom!=NULL)
    {
      elimLNombre(&nom);
    }
  return rc;  
}



/**
 * Test if a symbol is defined as a generator in the list that receives
 *
 *@param gen Generators list
 *@param symb Symbol 
 *
 *@return 0 iff it can't find the symbol
 **/
int existsGenerator(nodo_generadora *gen,char *symb)
{
  int exists=0;
  assert(symb!=NULL);
  
  while (!exists && gen!=NULL)
    {
      exists=(strcmp(gen->nombre,symb)==0);
      gen=gen->sig;
    }
  return exists;
}


/**
 * Test if a symbol is defined as a selector in the list that receives
 *
 *@param sel Selectors list
 *@param symb Symbol 
 *
 *@return 0 iff it can't find the symbol
 **/
int existsSelector(nodo_select *sel,char *symb)
{
  int exists=0;
  assert(symb!=NULL);
  
  while (!exists && sel!=NULL)
    {
      exists=(strcmp(sel->nombre,symb)==0);
      sel=sel->sig;
    }
  return exists;
}

/**
 * Test if a symbol is defined as a names list
 *
 *@param sel Names list
 *@param symb Symbol 
 *
 *@return 0 iff it can't find the symbol
 **/
int existsName(nodo_nombre *nom,char *symb)
{
  int exists=0;
  assert(symb!=NULL);
  
  while (!exists && nom!=NULL)
    {
      exists=(strcmp(nom->nombre,symb)==0);
      nom=nom->sig;
    }
  return exists;
}

	

/**
 * Test if a symbol is already defined in the symbol table
 *
 *@param symt Symbol Table
 *@param symb Symbol to look for
 *@param ntad TAD name where the symbol is defined
 *
 *@return 0 if it can't find the symbol name,
 *        SYM_ADT if it finds and ADT with the name
 *        SYM_INITAL if it finds a initial 
 *        SYM_CONSTRUCTOR if it finds a constuctor
 *        SYM_SELECTOR if it finds a selector
 **/
int existsSymbol(nodo_tad *symt,char *symb,char *ntad)
{
  int exists=0;

  while (!exists && symt!=NULL)
    {
      assert(symb!=NULL);
      assert(ntad!=NULL);
      /* Is the ADT name? */
      if (strcmp(symt->nombre,symb)==0)/* Is the ADT name? */
	{ exists=SYM_ADT; }
      else if (existsGenerator(symt->iniciales,symb)) /* Initial? */
	{ exists=SYM_INITIAL; }
      else if (existsGenerator(symt->constructoras,symb)) /* Constructor? */
	{ exists=SYM_CONSTRUCTOR; } 
      else if (existsSelector(symt->selectoras,symb)) /* Selector? */
	{ exists=SYM_SELECTOR; }
      else
	{ symt=symt->sig;}
    }

  if (exists!=0)
    { strcpy(ntad,symt->nombre); }
  return exists;
}


/**
 * Test if a parameter is already defined in the symbol table
 *
 *@param symt Symbol Table
 *@param par Symbol to look for
 *@param ntad TAD name where the parameter is defined
 *
 *@return 0 if it can't find the parameter name,
 *        1 if it finds it and returns de ADT in ntad
 **/
int existsSymbolParameter(nodo_tad *symt,char *par,char *ntad)
{
  int exists=0;

  while (!exists && symt!=NULL)
    {
      assert(par!=NULL);
      assert(ntad!=NULL);
      /* Is the ADT name? */
      if (existsName(symt->parametros,par)) /* Is a parameter of this ADT? */
	{
	  exists=1;
	  strcpy(ntad,symt->nombre);
	}
      symt=symt->sig;
    }

  return exists;
}



/**
 * Update the symbol table with the ADT that receives
 * (if it exists the records are updated, if it doesn't exists
 *  a new item is created in the symbol table)
 *
 *@param symt Symbols table
 *@param n ADT to update
 *@param merr Buffer to store error messages
 **/
int updateSymbols(nodo_tad **symt,nodo_tad *n,char *merr)
{
  nodo_tad **ant,*s;
  nodo_tad *nuevo;

  assert(symt!=NULL);
  assert(n!=NULL);
  assert(merr!=NULL);

  /* First we make a copy of the ADT signature */
  nuevo=NULL;
  if (copiaNodoTipo(&nuevo,n,merr)!=0)
    {
      return -1;
    }
  freeAxioms(nuevo->selectoras);

  /* Then we remove the older information (if exists) and replace (add)
     with the copy we made */
  for (ant=symt,s=*symt;s!=NULL && strcmp(s->nombre,n->nombre)!=0;
       ant=&(s->sig),s=s->sig);
  nuevo->sig=NULL;
  if (s!=NULL)
    {
      nuevo->sig=s->sig;
      elimNodoTipo(&s);
    }
  *ant=nuevo;
  return 0;
}


/**
 * Frees all the symbols of the symbol table that receives
 *
 * @param symt Symbol table
 **/
void freeSymbols(nodo_tad **symt)
{

  nodo_tad *ns;
  assert(symt!=NULL);

  while (*symt!=NULL)
    {
      ns=(*symt)->sig;
      elimNodoTipo(symt);
      symt=&ns;
    }
}


/**
 * Removes the symbols of the ADT that receives
 *
 * @param symt Symbol table
 **/
void removeSymbolADT(nodo_tad **symt,char *tn)
{
  nodo_tad *ac,**ant;
  
  assert(symt!=NULL);
  assert(tn!=NULL);

  ac=*symt;
  ant=symt;
  for (;ac!=NULL && strcmp(ac->nombre,tn);
         ant=&(ac->sig),ac = ac->sig)
    ;
  if (ac!=NULL) /* Encontro el nodo por eliminar */
    {
      *ant=ac->sig;
      ac->sig=NULL;
      elimNodoTipo(&ac);
    }
}

